
from pandas import DataFrame
from pandas.core.groupby import DataFrameGroupBy
from typing import Iterable
from pathlib import Path
from time import sleep

import requests

#TODO: maybe create an abstract base class and separate the offline and service implementations
class contributionsManager:
    def __init__(self, requestID, contributors_path : Path = None, endpoint="https://toar-data.fz-juelich.de/api/v2/request_contributors"):
        self.requestID = requestID
        self.timeseriesIDs = set()
        self.endpoint = endpoint
        self.runsAsService = True
        self.inline_mode = False
        if contributors_path is not None:
            self.runsAsService = False
            self.contributors_path = contributors_path
    def setup_contributors_endpoint_for_metadata(self):
        """!create the contributors endpoint depending on the intended mode.

        This either adds the timeseries IDs to the database and provides a get endpoint for the contributors or writes the timeseries IDs to a file and provides a curl command to upload the file to the database.
        """
        if self.runsAsService:
            return self.setup_contributors_service()
        else:
            if self.inline_mode:
                return self.setup_contributors_inline()
            else:
                return self.setup_contributors_id_file()
    def setup_contributors_inline(self) -> str:
        return f"{self.endpoint}" + "/?timeseriesids=" + ",".join([str(id) for id in self.timeseriesIDs])
    def setup_contributors_id_file(self) -> str:
        ext = "contributors"
        with open(self.contributors_path / f"{self.requestID}.{ext}", "w") as f:
            for id in self.timeseriesIDs:
                f.write(f"{id}\n")
        return f'curl -X POST "file=@{self.requestID}.{ext}" {self.endpoint}'
    def setup_contributors_service(self) -> str:
        # TODO: missing implementation
        raise NotImplementedError("This has not been implemented as this package is not yet operated as a service.")
    def add_timeseries_id(self, id : int) -> None:
        self.timeseriesIDs.add(id)
    def add_timeseries_ids(self, ids : Iterable[int]) -> None:
        for id in ids:
            self.timeseriesIDs.add(id)
    def extract_contributors_from_data_frame(self, data_frame : DataFrame):
        self.add_timeseries_ids( data_frame.index.to_list() )
    def extract_contributors_from_grouped_dataframe(self, data_grouped_by_cell : DataFrameGroupBy):
        for _, table in data_grouped_by_cell:
            self.add_timeseries_ids( table.index.to_list() )

class contributions_manager_by_name(contributionsManager):
    def add_timeseries_id(self, id : int) -> None:
        names = self.id_to_names(id)
        for name in names:
            self.timeseriesIDs.add(name)
    def add_timeseries_ids(self, ids : Iterable[int]) -> None:
        for id in ids:
            names = self.id_to_names(id)
            for name in names:
                self.timeseriesIDs.add(name)
    def id_to_names(self, id : int) -> list[str]:
        for pos in range(10):
            req_res = requests.get(f"https://toar-data.fz-juelich.de/api/v2/timeseries/{id}", timeout=40)
            try:
                results = req_res.json()
                break
            except:
                print("Test debug:", pos, " try for id", id)
                sleep(30)
                pass
        else:
            raise RuntimeError(f"Could not get the response for the timeseries with id {id}.")
        names = set()
        for r in results["roles"]:
            try:
                #print("\t", r["contact"]["person"]["name"])
                names.add( r["contact"]["person"]["name"] )
            except:
                try:
                    #print("\t", r["contact"]["organisation"]["name"])
                    names.add( r["contact"]["organisation"]["name"] )
                except:
                    raise RuntimeError("Could not find 'person' or 'organisation' in the response.\n" + str(r))
        return names
    def setup_contributors_inline(self) -> str:
        return ", ".join([str(id) for id in sorted(self.timeseriesIDs)])
    def setup_contributors_id_file(self) -> str:
        ext = "contributors"
        with open(self.contributors_path / f"{self.requestID}.{ext}", "w") as f:
            for id in self.timeseriesIDs:
                f.write(f"{id}\n")
        return f'see file {self.requestID}.{ext}"'

