import logging
import sys
from collections import namedtuple
from logging.handlers import SysLogHandler, TimedRotatingFileHandler
from pathlib import Path

handlerPair = namedtuple("registeredLogger", ["handler", "formatter"])


class toargridding_defaultLogging:
    """! class to setup default loggers for toargridding

    The added handler and their formatters are stored by this class for further configuration.
    Parameters:
    ----------
    loggername:
        name of the logger to be used. default: toargridding

    Methods:
    -------
    registerHandler:
        register a handler. Adds it to the logger and stores references in this class
    addShellLogger:
        adds a formatting logging to the shell. Default level is DEBUG
    addSysLogger:
        adds a formatted logging to the system log (on a linux system). Default level is WARNING
    logExceptions:
        enable passing of any uncaught exception to the logger
    getHandler:
        get a registered handler
    getFormatter:
        get a registered formatter
    """

    def __init__(self, loggername: str = "toargridding"):
        self.logger = logging.getLogger(loggername)
        self.logger.setLevel(logging.DEBUG)
        self.registeredHandlers = {}  # name : registeredLogger

    def registerHandler(self, name: str, handler, formatter=None):
        """register a handler. Adds it to the logger and stores references in this class.
        The given formatter (if not None) is added to the handler.
        The handler is added to the logger.
        Handler and logger are stored for a possible later access.

        Throws an exception if the name of the handler is already known.
        """
        if name in self.registeredHandlers:
            msg = f"There is already a registered handler with the name {name}."
            raise ValueError(msg)
        if formatter is not None:
            handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        self.registeredHandlers[name] = handlerPair(handler, formatter)

    def getHandler(self, name: str):
        """get a handler for logging by its name used for the registration"""
        if name in self.registeredHandlers:
            return self.registeredHandlers[name].handler
        return None

    def getFormatter(self, name: str):
        """get a formatter by its name used for the registration"""
        if name in self.registeredHandlers:
            return self.registeredHandlers[name].formatter
        return None

    def addShellLogger(self, level=logging.INFO):
        """!adds a formatted logging to the shell to the "toargridding" logger.
        The handler is registered as "shell".

        Parameters:
        ----------
        level:
            set the verbosity level of the logger (default: info)
        """
        shell_handler = logging.StreamHandler()
        shell_handler.setLevel(level)
        shell_formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s] - %(filename)s:%(lineno)d: '%(message)s'", datefmt="%Y-%m-%d %H:%M:%S"
        )
        self.registerHandler("shell", shell_handler, shell_formatter)

    def addSysLogger(self, level=logging.WARNING):
        """!adds a formatted logging to the system log of a linux system to the "toargridding" logger.
        This logging is registered as "syslog".

        Parameters:
        ----------
        level:
            set the verbosity level of the logger (default: warning)
        """
        syslog_handler = SysLogHandler(facility=SysLogHandler.LOG_DAEMON, address="/dev/log")
        syslog_formatter = logging.Formatter(
            fmt="TOARGRIDDING [%(levelname)s] - %(filename)s:%(lineno)d: '%(message)s'"
        )
        syslog_handler.setLevel(level)
        self.registerHandler("syslog", syslog_handler, syslog_formatter)

    def addRotatingLogFile(self, filename: Path, level=logging.INFO):
        """creation of a rotating file handler, that will change the files at midnight.
        The last 7 files logfiles are stored.

        Parameters:
        ----------
        filename:
            basename of the file. The parent path will be created, if required.
        level:
            output level for this handler
        """
        filename.parent.mkdir(parents=True, exist_ok=True)
        handler = TimedRotatingFileHandler(filename, when="midnight", backupCount=7)
        handler.setLevel(level)
        formatter = logging.Formatter(
            fmt="%(asctime)s [%(levelname)s] - %(filename)s:%(lineno)d: '%(message)s'", datefmt="%Y-%m-%d %H:%M:%S"
        )
        self.registerHandler("rotatingFile", handler, formatter)

    def addRotatingLogFile_scriptName(self, scriptName: str | Path, level=logging.INFO):
        """creation of an rotating log file by using the script name.
        In the /path/to/script a subdirectory log will be created. The logfile will be name [script basename].log

        Parameters:
        ----------
        scriptName:
            name to the script, including its path
        level:
            verbosity level.
        """
        sn = Path(scriptName)
        if not sn.is_file:
            msg = f"Expecting name to a script. {sn} is not a file."
            raise ValueError(msg)
        path = sn.parent / "log"
        path.mkdir(exist_ok=True)
        fn = path / f"{sn.stem}.log"
        self.addRotatingLogFile(fn, level)

    def logExceptions(self):
        """calling this function will redirect all uncaught exceptions to the logger
        This is especially useful to write the exceptions to the system log
        """
        sys.excepthook = self.handle_exception

    def handle_exception(self, exc_type, exc_value, exc_traceback):
        """function for passing uncaught exceptions to the logger"""
        if issubclass(exc_type, KeyboardInterrupt):
            sys.__excepthook__(exc_type, exc_value, exc_traceback)
            return

        self.logger.error(
            "Program terminated by the following exception:", exc_info=(exc_type, exc_value, exc_traceback)
        )
