from collections import namedtuple
from itertools import product
from pathlib import Path

import xarray as xr

from toargridding.grids import GridDefinition
from toargridding.metadata import Metadata, TimeSample
from toargridding.toar_rest_client import AnalysisService

from toargridding.contributors import contributionsManager

GriddedResult = namedtuple("GriddedResult", ["dataset", "metadata"])


def get_gridded_toar_data(
    analysis_service: AnalysisService,
    grid: GridDefinition,
    time: TimeSample,
    variables: list[str],
    stats: list[str],
    contributors_path : Path = None,
    **kwargs,
) -> tuple[list[xr.Dataset], list[Metadata]]:
    """API to download data as xarrays

    The function creates all combinations of the variable and stats list

    Parameters:
    ----------
    analysis_service:
        access to the REST API and manages download of data
    grid:
        grid for the output data.
    time:
        sampled values in time
    variables:
        list of variables to be extracted from the TOARdatabase
    stats:
        list of statistical properties to be extracted from the TOAR database
    contributors_path:
        pathname to write the contributors path. We advise to store the contributor files into the same directory as the resulting data.
        Without a provided path, it is assumed, that toargridding is operated as a service and the contributors can be directly provided through the contributors endpoint. This is not yet implemented.

    return:
    -------
        Gridded datasets for each combination of variables and stats and appropriate metadata for each dataset.
    """

    metadatas = [
        Metadata.construct(standard_name=var, time=time, stat=stat, moreOptions=kwargs)
        for var, stat in product(variables, stats)
    ]

    datasets = []
    for metadata in metadatas:  # standard_name ?
        data = analysis_service.get_data(metadata)
        #TODO add processing of contributors
        # create contributors endpoint and write result to metadata
        contributors_field = contributionsManager(metadata.get_id(), contributors_path)
        contributors_field.extract_contributors_from_data_frame(data.stations_data)
        metadata.contributors_metadata_field = contributors_field.setup_contributors_endpoint_for_metadata()
        ds = grid.as_xarray(data)
        datasets.append(ds)

    # TODO: return this as a list of tuples to keep data and metadata together?
    return datasets, metadatas
