import logging
from abc import ABC, abstractmethod
from collections import namedtuple
from enum import Enum

import numpy as np
import pandas as pd
import xarray as xr
from pandas.core.groupby import DataFrameGroupBy

from toargridding.metadata import (
    AnalysisRequestResult,
    Coordinates,
    Metadata,
    Variables,
    get_global_attributes,
)
from toargridding.variables import Coordinate, Variable
from toargridding.contributors import contributionsManager
from sys import exit

logger = logging.getLogger(__name__)

GridType = Enum("GridType", ["regular"])
"""list of available grids.
"""


class GridDefinition(ABC):
    """factory and base class for definition of different grids

    usage: GridDefinition.construct( GridType, dict( parameter : value ) )
    The dict must contain all parameters required for the creation of the desired GridType
    """

    cell_index_name = "cell_index"

    def __init__(self):
        self.fill_value = np.nan  # TO(pd.DataFrame)) -> xr.Dataset:
        # TODO make CF-compliant

    @staticmethod
    def construct(grid_type: GridType, **kwargs):
        """creation of requested grid type

        usage: GridDefinition.construct( GridType, dict( parameter : value ) )
        The dict must contain all parameters required for the creation of the desired GridType
        """
        match grid_type:
            case GridType.regular:
                return RegularGrid(**kwargs)
            case other:
                raise Exception("invalid grid type")

    @property
    @abstractmethod
    def description(self) -> str:
        """description of this grid"""

    @staticmethod
    def from_netcdf():
        """loading of an arbitrary grid from a netCDF file

        not yet implemented
        """

    @abstractmethod
    def as_xarray(self, data: AnalysisRequestResult) -> xr.Dataset:
        """gridding of a request to the TOAR database

        This includes the required setup to store the results as netCDF file according to CF (https://cfconventions.org/cf-conventions/cf-conventions.html)
        """

    @staticmethod
    def get_id():
        """provide an ID for this grid and its resolution."""


class RegularGrid(GridDefinition):
    """definition of a regular grid with longitude and latitude.

    The grid covers the complete globe and is defined by providing resolution for latitude (lat_resolution) and longitude (lon_resolution)

    Argument:
    --------
    lat:
        latitude coordinate axis.
    lon:
        longitude coordinate axis.
    dims:
        names of the dimensions of the data
    """

    Coord = namedtuple("Coord", ["lat", "lon"])

    def __init__(self, lat_resolution, lon_resolution):
        """constructor from resolutions

        Parameters:
        ----------
        lat_resolution:
            resolution for latitude in degree
        lon_resolution:
            resolution for longitude in degree
        """

        super().__init__()
        # TODO make sure only sensible resolutions

        self.lat = Coordinate.from_resolution(Coordinates.latitude, lat_resolution, min=-90, max=90, wraps=False)
        self.lon = Coordinate.from_resolution(Coordinates.longitude, lon_resolution, min=-180, max=180, wraps=True)
        spatial_shape = (self.lon.size, self.lat.size)
        spatial_size = self.lon.size * self.lat.size
        self.dims = [
            Coordinates.time.name,
            Coordinates.latitude.name,
            Coordinates.longitude.name,
        ]

        self._as_xy_index = np.dstack(np.meshgrid(range(self.lat.size), range(self.lon.size))).reshape(-1, 2)
        self._as_i_index = np.arange(spatial_size).reshape(spatial_shape).T

    @property
    def description(self) -> str:
        """get description of grid"""

        return f"regular global grid with lat/lon resolutions ({self.lat.step}, {self.lon.step})"

    def as_xarray(self, data: AnalysisRequestResult) -> xr.Dataset:
        """gridding of a request to the TOAR database

        groups the stations into the cells of the grid and calculates mean and standard deviation for each cell.

        Parameters:
        ----------
        data:
            results of the request, including data, station coordinates and metadata of request
        """
        data_grouped_by_cell = self.group_data_by_cell(data.stations_data, data.stations_coords)
        cell_statistics = self.get_cell_statistics(data_grouped_by_cell)
        dataset = self.create_dataset(cell_statistics, data.metadata)

        return dataset

    def group_data_by_cell(self, data: pd.DataFrame, coords: pd.DataFrame) -> DataFrameGroupBy:
        """grouping of stations into cells

        This function converts the lat/lon coordinates of the stations into cell indices and groups stations belonging to one cell.

        Parameters:
        ----------
        data:
            station data
        coords:
            station coordinates
        """

        cell_indices = self.as_cell_index(coords)

        # will convert cell_indices to float as some nans ar present
        data_with_indices = data.join(cell_indices.to_frame(GridDefinition.cell_index_name), how="inner")

        return data_with_indices.groupby(GridDefinition.cell_index_name)

    def get_cell_statistics(self, groups: DataFrameGroupBy) -> dict[str, pd.DataFrame]:
        """calculation of mean, std and number of stations per cell

        Parameters:
        ----------
        groups:
            time series data grouped by stations in a cell
        return:
            dictionary with calculated quantities
        """

        stats = {
            Variables.mean: groups.mean(),
            Variables.std: groups.std(),
            Variables.n: groups.count(),
        }

        return stats

    def create_dataset(self, cell_statistics: dict[str, pd.DataFrame], metadata: Metadata) -> xr.Dataset:
        """creation of data set and filling with results from the gridding

        Parameters:
        ----------
        cell_statistics:
            values obtained for each cell. In the beginning these are mean, std and number of stations
        metadata:
            metadata of the request
        return:
            xarray dataset with coordinates and variables following the CF convention
        """

        time = Coordinate.from_data(
            metadata.time.as_cf_index(cell_statistics[Variables.mean].columns),
            Coordinates.time,
            metadata,
            step=metadata.time.sampling,
        )

        gridded_ds = self.get_empty_grid(time, metadata)
        for variable, aggregated_data in cell_statistics.items():
            data_array_dict = self.get_data_array_dict(time, aggregated_data, variable, metadata)
            gridded_ds = gridded_ds.assign(data_array_dict)

        return gridded_ds

    def get_data_array_dict(
        self,
        time: Coordinate,
        aggregated_data: pd.DataFrame,
        variable: Variables,
        metadata: Metadata,
    ) -> dict[str, xr.DataArray]:
        """conversion of data to a dict for assigning them to the Dataset

        Parameters:
        ----------
        time:
            temporal coordinate
        aggregated_data:
            obtained results per grid cell
        variable:
            variable to be added. This allows access to Cf conform metadata
        metadata:
            metadata of request
        """

        gridded_data = self.create_gridded_data(time, aggregated_data)
        gridded_variable = Variable.from_data(gridded_data, variable, metadata)
        return {variable.name: gridded_variable.as_data_array(self.dims)}

    def create_gridded_data(self, time: Coordinate, grouped_timeseries: pd.DataFrame) -> np.array:
        """converts the available cell data to a full lat/lon-temporal data cube.

        Parameters:
        ----------
        time:
            temporal coordinate
        grouped_timeseries:
            data frame with station position and data
        return:
            3D-array with axis time, latitude and longitude. Fields without data are nan (fill_value defined in GridDefinition init)
        """

        # CAVE: This function might involve black magic...
        values = np.empty((time.size, self.lat.size, self.lon.size))
        values[...] = self.fill_value

        index = self._as_xy_index[grouped_timeseries.index.astype(int)]
        values[:, index.T[0], index.T[1]] = grouped_timeseries.values.reshape(-1, time.size).T

        return values

    def as_cell_index(self, coords: pd.DataFrame) -> pd.Series:
        """converts coordinates of stations into x and y indices of the regular grid"""

        id_x = self.coord_to_index(coords[self.lat.name], self.lat.min, self.lat.step)
        id_y = self.coord_to_index(coords[self.lon.name], self.lon.min, self.lon.step, len(self.lon.data))

        id_i = self._as_i_index[id_x, id_y]

        return pd.Series(id_i, index=id_x.index)

    def coord_to_index(self, coord: pd.Series, x0_axis: float, d_axis: float, maxBin4Wrap: int = None) -> np.array:
        """converts a coordinate into a bin index on one axis

        Parameters:
        ----------
        coord:
            coordinate for conversion
        x0_axis:
            offset of the axis
        d_axis:
            resolution of the axis
        maxBin4Wrap:
            this is the maximum bin number. If it is set, thw bin numbers wrap around, like for the latitude
        """
        ids = (np.ceil((coord / d_axis) - 0.5) - x0_axis / d_axis).astype(int)
        if maxBin4Wrap is not None:
            ids[ids >= maxBin4Wrap] %= maxBin4Wrap
            ids[ids < 0] += maxBin4Wrap
        return ids

    def get_empty_grid(self, time: Variable, metadata: Metadata) -> xr.Dataset:  # TODO make CF-compliant => docs
        """creation of an empty dataset without data

        Sets up a dataset with its three axis: time, longitude and latitude.
        Adds global metadata to the dataset.

        Parameters:
        ----------
        time:
            temporal coordinate
        Metadata:
            information on request
        """

        coords = {
            Variables.time.name: time.as_data_array(),
            Variables.latitude.name: self.lat.as_data_array(),
            Variables.longitude.name: self.lon.as_data_array(),
        }

        ds = xr.Dataset(coords=coords, attrs=get_global_attributes(metadata))

        return ds

    def get_id(self):
        """provide an ID for this grid and its resolution."""
        return f"_regular{len(self.lat.data)}x{len(self.lon.data)}"
