import io
import json
import logging
import time
from contextlib import contextmanager
from dataclasses import asdict, dataclass, field
from datetime import datetime
from pathlib import Path
from zipfile import ZipFile

import pandas as pd
import requests

from toargridding.metadata import AnalysisRequestResult, Coordinates, Metadata
from toargridding.contributors import contributionsManager

logger = logging.getLogger(__name__)

STATION_LAT = "station_coordinates_lat"
STATION_LON = "station_coordinates_lng"
COORDS = [STATION_LAT, STATION_LON]


class EmptyDataError(ValueError):
    """! custom exception for requests, where the analysis service only provides metadata.

    This might be the case, if there are not statuins, or if the statistical analysis does not yield any data points
    """

    def __init__(self, message):
        super().__init__(message)


@dataclass(frozen=True)
class QueryOptions:
    """Creation of a request to the TOAR database.

    Attributes:
    ----------
    datarange:
        extraction of timestamps with start<=valid<end. Format: [start],[end] with start and end in isoformat
    variable_id:
        id of the variable within the TOAR database
    statistics:
        statistical quantity requested from the TOAR database. see toargridding.toarstats_constants.STATISTICS_LIST.
    sampling:
        temporal aggregation, e.g. daily, monthly
    min_data_capture:
        most probably the minimum data to include in the request
    metadata_schema:
        amount of metadata being provided, see Quick Start for TOAR Analysis Service
    limit:
        limit to amount of extracted data; see Quick Start for TOAR Analysis Service
    format:
        output format; see Quick Start for TOAR Analysis Service
    moreOptions:
        dict with additional query options for the request to the TOAR database.
    """

    daterange: str
    variable_id: str
    statistics: str
    sampling: str
    min_data_capture: str = "0"  # TODO check effect on NaNs
    metadata_scheme: str = "basic"
    limit: str = "None"
    format: str = "by_statistic"
    moreOptions: dict = field(default_factory=dict)  # needs to be last element for to dict factory

    @staticmethod
    def from_metadata(metadata: Metadata):
        """Creation from Metadata object

        Copies datarange, variable_id, statistics and sampling from the metadata object.
        For the other parameters the default values are used.
        """
        return QueryOptions(
            daterange=metadata.time.daterange_option,
            variable_id=str(metadata.variable.toar_id),
            statistics=metadata.statistic,
            sampling=metadata.time.sampling,
            moreOptions=metadata.moreOptions,
        )

    @property
    def cache_key(self):
        """creation to identify the request in the cache of known request."""
        return "".join(f"{key}{val}" for key, val in sorted(asdict(self, dict_factory=quarryToDict).items()))


def quarryToDict(data: QueryOptions):
    out = dict(data[:-1])
    extraVals = data[-1][1]
    for field, value in extraVals.items():
        if field not in data:
            out[field] = value
        else:
            msg = f"Providing invalid value for TAOR database: {field} is controlled by Metadata class"
            raise ValueError(msg)
    return out


class Cache:
    """cache to store download links for requests to the TOAD database

    The extraction of data and statistical processing by the TOAR database can take hours or even days, depending on the amount of requested data.
    The cache is persistent to allow the access to data after turning of the computer.

    It is created in a dict like way and supports operations like "in".
    A textfile with the extension "_v2.json" is created in cache_dir. CAVE: The first entry is required for loading the empty file.
    The cache also saves the creation time of a status endpoint and automatically deletes entries that are to old. The maximum age in days is set via Cache.setMaxDaysInCache().

    """

    maxDaysInCache = 14
    __timeFormat = "%Y%m%d-%H:%M"

    def __init__(self, cache_dir: Path, fn="status_endpoints"):
        """constructor

        Throws exception if cache directory does not exists.
        Parameters
        ----------
        cache_dir:
            directory for storing cache file.
        """

        if not cache_dir.exists():
            msg = f"Given directory for saving cache file does not exists. Path: {cache_dir}"
            raise RuntimeError(msg)
        self.cache_file = cache_dir / f"{fn}_v2.json"

        if not self.cache_file.is_file():  # initialize cache with dummy values
            with open(self.cache_file, "w") as cache:
                json.dump({"foo": ["bar", self.__getTime()]}, cache)

    def __contains__(self, item: str):
        """allows usage of "in" """
        with self.storage_dict() as storage:
            return item in storage

    def get(self, key: str):
        """get an endpoint from the cache."""
        with self.storage_dict() as storage:
            return storage[key][0]

    def put(self, key: str, content: str):
        """get add key and content as key-value-pair to cache"""
        with self.storage_dict() as storage:
            storage[key] = [content, self.__getTime()]

    def remove(self, key: str):
        """remove a key and content as key-value-pair to cache"""
        with self.storage_dict() as storage:
            del storage[key]

    @contextmanager
    def storage_dict(self):
        with open(self.cache_file) as cache:  # setup
            storage_dict = json.load(cache)

        yield storage_dict

        with open(self.cache_file, "w") as cache:  # teardown
            json.dump(storage_dict, cache, indent=2)

    def __getTime(self):
        """! get now as string for saving in cache"""
        return datetime.now().strftime(Cache.__timeFormat)

    def clear_cache(self):
        """!Delete all values from the cache, that are older than the maxDaysInCache value"""
        with self.storage_dict() as storage:
            now = datetime.now()
            storage["foo"] = ["bar", self.__getTime()]  # preserve first line to ensure, that the file can be read
            deleteList = []
            for key, [_, creation] in storage.items():
                creationTime = datetime.strptime(creation, Cache.__timeFormat)
                age = now - creationTime
                age = float(age.days) + age.seconds / 24 / 3600.0
                if age > Cache.maxDaysInCache:
                    deleteList.append(key)
            for key in deleteList:
                del storage[key]

    @staticmethod
    def set_max_days_in_cache(maxDays: float):
        """! set the maximum age of entries in the cache.
        Also accepts fractions of a day down to seconds.
        Parameters:
        ----------
        maxDays:
            duration in days.
        """
        if maxDays > 0:
            Cache.maxDaysInCache = maxDays
        else:
            msg = f"The maximum age needs to be larger than 0. You provided {maxDays}."
            raise ValueError(msg)


class Connection:
    def __init__(self, endpoint, cache_dir):
        """connection to the rest API of the TOAR database

        This class handles the creation of requests and the interaction with the TOAR database.
        It relies on the "TOARDB Analysis FastAPI REST interface"
        https://toar-data.fz-juelich.de/api/v2/analysis/
        with the default endpoint: https://toar-data.fz-juelich.de/api/v2/analysis/statistics

        Parameters:
        ----------
        endpoint:
            link to the TOAR database analysis service
        cache_dir:
            directory to store cache file
        """

        self.endpoint = endpoint
        self.cache = Cache(cache_dir)
        self.cache.clear_cache()
        self.cache_backup = Cache(cache_dir, "status_endpoints.old")
        # max wait time is 30min
        self.wait_seconds = []
        self.set_request_times(5, 30)

    def set_request_times(self, interval_min, max_wait_minutes):
        """set the intervals and maximum duration to wait, before requests to the analysis service are stopped

        The waiting intervals determine how long and often the status endpoint is checked if the results are available.
        As soon as the maximum waiting time is reached, the process is stopped. It can be restarted at any time, as the required endpoint is stored in a cache file

        Parameters:
        -----------
        interval_min:
            waiting interval in minutes
        maxWait_min:
            maximum duration to wait in minutes.
        """
        if max_wait_minutes <= 0:
            msg = "The maximum waiting time needs to be larger than 0min."
            raise RuntimeError(msg)
        elif interval_min < 0 or interval_min > max_wait_minutes:
            self.wait_seconds[0] = max_wait_minutes
        else:
            self.wait_seconds = [interval_min * 60 for _ in range(interval_min, max_wait_minutes + 1, interval_min)]

    def get(self, query_options: QueryOptions) -> requests.models.Response:
        """get results for a request.

        This is the main function to obtained data from the TOAR DB. It will start requests or lookup if an already started requests is finished.
        As soon as a request yields an status code 500 (which most likely indicates, that the worker of the analysis service has crashed), the endpoint is discarded and the request is resubmitted.

        Throws an exception, if the results are not available after the waiting time. A restart of the function continues the regular lookup for results.
        This function catches possible connection issues and continues to
        """
        status_endpoint = self.get_status_endpoint(query_options)
        nRestart = 0
        waitRestart = 300
        for i, wait_time in enumerate(self.wait_seconds):
            logger.info(f"try: {i+1}, wait_time: {wait_time}")
            response = self.wait_and_get(status_endpoint, wait_secs=wait_time)
            # do error handling i.e. look for connection issues
            try:
                response.raise_for_status()
            except requests.exceptions.HTTPError as e:
                logger.warning(f"\tconnection error ({e.response.status_code}: {e.response.reason}).")
                self.printExecption(e, response)
                # a Status Code 500 seems indicated an aborted request -> restart the request and continue with new status endpoint
                if e.response.status_code == 500:
                    logger.warning(f"Received internal server error. Restarting request in {waitRestart}s.")
                    time.sleep(waitRestart)
                    self.cache.remove(query_options.cache_key)
                    status_endpoint = self.get_status_endpoint(query_options)
                else:
                    logger.warning("\tTrying again later.")
                continue
            # are our results ready to obtain?
            if response.headers["Content-Type"] == "application/zip":
                logger.info("Results are available for download")
                return response
        msg = f"No data available after {sum(self.wait_seconds) / 60 + nRestart*waitRestart / 60} minutes. retry later."
        raise RuntimeError(msg)

    def get_status_endpoint(self, query_options: QueryOptions) -> str:
        """get endpoint to results of a request

        This function checks if the request is already known and has been submitted to the TOAD DB.
        If yes, the know endpoint is returned.
        If the cache knows the endpoint, but the DB has deleted it, the endpoint is removed from the cache and a new request is started.
        Otherwise a new new request is started.

        Throws an RuntimeError in case, of a connection error or any other error. In case of an HTTPError, the request is removed from the cache

        Parameters:
        ----------
        Options for the request.
        """
        if query_options.cache_key in self.cache:
            status_endpoint = self.cache.get(query_options.cache_key)

            try:  # test for stale cache
                response = self.wait_and_get(status_endpoint)
                response.raise_for_status()
            except requests.exceptions.ReadTimeout:
                logger.critical("Caught read timeout.")
                msg = "Connection to TAORDB timed out (ReadTimeout) while checking cached status point. Please try again later."
                raise RuntimeError(msg)
            except requests.exceptions.HTTPError as e:
                # TODO add detailed processing: What was the reason for the error? Do we really need to create a new request or is there another problem, that might resolve by simply waiting
                logger.warning("A HTTP error occurred:")
                self.printExecption(e, response)
                logger.debug(f"Status Endpoint: {status_endpoint}")
                # use inverse order for saving. the status endpoint should be more unique
                self.cache_backup.put(status_endpoint, query_options.cache_key)
                # will be overwritten in the next step.
                self.cache.remove(query_options.cache_key)
                logger.warning("Removing status endpoint from cache and submitting new request.")
            except:
                msg = "An error occurred during accessing a cached request"
                raise RuntimeError(msg)
            else:
                logger.info("load status endpoint from cache")
                return status_endpoint
        else:
            logger.info("query not in cache")

        return self.query_for_status_endpoint(query_options)

    def query_for_status_endpoint(self, query_options: QueryOptions) -> str:
        """create and new request to the TOAR DB.

        Adds the status endpoint of the request to the cache.

        Throws an exception if the TOAR Db returns an error.

        Parameters:
        ----------
        query_options:
            request to the TOAR database.
        """
        try:
            response = self.wait_and_get(self.endpoint, asdict(query_options, dict_factory=quarryToDict))
            url = response.history[0].url if response.history else response.url
            logger.debug(f"Original request: {url}")

            if response.headers["Content-Type"] == "application/json":
                status_endpoint = response.json()["status"]
            response.raise_for_status()
        except requests.exceptions.HTTPError as e:
            logger.warning("An HTTP error occurred:")
            self.printExecption(e, response)
            raise
        except requests.exceptions.ReadTimeout as e:
            logger.critical("Caught read timeout.")
            self.printExecption(e, response)
            msg = "Read timeout while querying for status endpoint"
            raise RuntimeError(msg)
        except:
            msg = f"Request was not successful. Response by TOAR database: {response.text}"
            raise RuntimeError(msg)

        # we mage it: let's remember the status endpoint to get our results later:-)
        self.cache.put(query_options.cache_key, status_endpoint)

        return status_endpoint

    def wait_and_get(
        self, endpoint: str, query_options: dict | None = None, wait_secs=None, timeout=(3.05, 20)
    ) -> requests.models.Response:
        """accesses given endpoint

        Parameters:
        ----------
        endpoint:
            either full endpoint of a request of base endpoint.
        query_options:
            used with the base endpoint to create a request. If None, endpoint is expected to be a full endpoint
        wait_secs:
            sleep in seconds before starting request to TAORDB
        timeout:
            timeout for the request.
        """
        if wait_secs:
            time.sleep(wait_secs)

        return requests.get(endpoint, params=query_options, timeout=timeout)

    def printExecption(self, e: requests.exceptions.HTTPError, response: requests.Response):
        """!output different infos from an exception and the corresponding response."""
        logger.debug(f"Status Code: {e.response.status_code}")
        logger.debug(f"Reason: {e.response.reason}")
        logger.debug(f"Text: {e.response.text}")
        logger.debug(f"{response=}")
        logger.debug(f"{response.content=}")
        try:
            logger.debug(response.json())
        except Exception as e:
            logger.debug("Decoding as json failed.")


class AnalysisService:
    METADATA = "metadata"

    def __init__(self, stats_endpoint, cache_dir):
        """constructor for setting up cache and connection

        Parameters:
        ---------
        stats_endpoint:
            link to statistics service of TOAR DB
        cache_dir:
            directory to store cache file for requests, needs to exist
        """
        self.connection = Connection(stats_endpoint, cache_dir)

    def get_data(self, metadata: Metadata) -> AnalysisRequestResult:
        """main function to obtain data from the TOAR DB

        Handles requesting and loading of data into memory as soon as they are available.
        In addition the data and coordinates undergo a cleanup.
        This results in a datasets which only contain timeseries and station coordinates, if both data are present and valid, i.e. not nan.
        A timeseries needs to have at least 1 non nan value to be considered as valid.

        Parameters:
        ----------
        metadata:
            meta data for the request.
        return:
            Requested data and statistics, station coordinates and metadata of the request.
        """

        timeseries, timeseries_metadata = self.get_timeseries_and_metadata(metadata)
        coords, timeseries = self.get_clean_timeseries_and_data(timeseries, timeseries_metadata, metadata )
        return AnalysisRequestResult(timeseries, coords, metadata)

    def get_timeseries_and_metadata(self, metadata: Metadata) -> tuple[pd.DataFrame, pd.DataFrame]:
        """obtain data and metadata from TOAR database

        return:
            tuple[timeseries, station coordinates]
        """

        query_options = QueryOptions.from_metadata(metadata)
        result = self.connection.get(query_options)
        timeseries, timeseries_metadata = self.load_data(result.content, metadata)
        return timeseries, timeseries_metadata
    
    def get_clean_timeseries_and_data(self, timeseries : pd.DataFrame, timeseries_metadata : pd.DataFrame, metadata : Metadata) -> tuple[pd.DataFrame, pd.DataFrame]:
        coords = self.get_clean_coords(timeseries_metadata)
        timeseries = self.get_clean_timeseries(timeseries, metadata)

        timeseries, coords = timeseries.align( coords, join="inner", axis=0 )
        return coords, timeseries


    def get_clean_coords(self, timeseries_metadata: pd.DataFrame):
        """remove all stations with invalid coordinates and drop unused metadata
        invalid coordinates are NaN, none etc.
        return:
            stations with valid coordinates
        """
        coords = timeseries_metadata[COORDS]
        coords.columns = [Coordinates.latitude.name, Coordinates.longitude.name]
        valid_coords = coords.notna().all(axis=1)
        return coords[valid_coords]

    def get_clean_timeseries(self, timeseries: pd.DataFrame, metadata: Metadata):
        """remove all timeseries without any data from the results

        timeseries:
            extracted time series
        metadata:
            metadata belonging ot the timeseries.

        return:
            timeseries without empty timeseries
        """
        # TODO maybe use cf-index here already ?
        ##here we observe some differences in the number of timestamps.
        # remove data where utc -> sun/local ? time conversion leads to date shift

        # conversion from string to datetime objects for easier access
        # TODO maybe check if the conversion is also valid for annual data
        timeseries.columns = pd.DatetimeIndex([datetime.strptime(val, "%Y-%m-%d") for val in timeseries.columns])

        # now drop columns outside of our requested range:
        # this is independent of the other part
        col2Drop = []
        for retrievedDate in timeseries.columns:
            if retrievedDate < metadata.time.start:
                col2Drop.append(retrievedDate)
            else:
                break
        for i in range(-1, -len(timeseries.columns), -1):
            retrievedDate = timeseries.columns[i]
            if retrievedDate > metadata.time.end:
                col2Drop.append(retrievedDate)
            else:
                break

        if len(col2Drop):
            logger.info(
                f"Dropping columns ({col2Drop}) from TOAR data to match requested date range [{metadata.time.start}, {metadata.time.end}]"
            )
            timeseries.drop(columns=col2Drop, inplace=True)

        all_na = timeseries.isna().all(axis=1)
        timeseries = timeseries[~all_na]
        return timeseries#.fillna(nan)

    def load_data(self, content: bytes, metadata: Metadata) -> tuple[pd.DataFrame, pd.DataFrame]:
        """convert downloaded byte stream into pandas dataframes

        throws an EmptyDataError, if the results file does not contain data.
        This is a result if there are not stations contributing to a request or if the restrictions of the analysis exclude all points of a station.

        Parameters:
        ----------
        content:
            downloaded data as zip file
        metadata:
            information on the request.
        """
        zip_stream = io.BytesIO(content)
        with ZipFile(zip_stream) as myzip:
            if len(myzip.namelist()) == 1:
                msg = "Data file from TOAR analysis service is empty"
                raise EmptyDataError(msg)
            timeseries = self.extract_data(myzip, metadata.statistic)
            timeseries_metadata = self.extract_data(myzip, AnalysisService.METADATA)

        return timeseries, timeseries_metadata

    def extract_data(self, zip_file: ZipFile, data_file: str) -> pd.DataFrame:
        """extract a specific csv file from the zip file

        Parameters:
        ----------
        zip_file:
            opened zip file
        data_file:
            base file name of the requested file. Extension .csv is added by this function.
        """
        with zip_file.open(f"{data_file}.csv") as f:
            s_stream = io.StringIO(f.read().decode("utf-8"))
            return pd.read_csv(s_stream, comment="#", index_col=0)


class AnalysisServiceDownload(AnalysisService):
    """download service with caching of requests to the TOARDB

    This service performs the request to the TOAR database and downloads the results of the request to disc before returning if for further processing.
    When retrieving data, a check is done, if this request has already been cached on disc.

    Attributes:
    ----------
    stats_endpoint:
        link to statistics service of TOAR DB
    cache_dir:
        directory to store cache file for requests, needs to exist
    sample_dir:
        directory for caching results of request to the TOARDB
    use_downloaded:
        flag to control if the cache of downloaded requests is checked before extracting data from the TOARDB
    """

    def __init__(self, stats_endpoint, cache_dir, sample_dir: Path, use_downloaded=True):
        super().__init__(stats_endpoint, cache_dir)
        self.sample_dir = sample_dir
        self.use_downloaded = use_downloaded

    def get_timeseries_and_metadata(self, metadata: Metadata) -> tuple[pd.DataFrame, pd.DataFrame]:
        """loading of cached data or requesting of data from the TOARDB

        Parameters:
        ----------
        metadata:
            information on requested data
        return:
            tuple[timeseries, station coordinates]
        """
        filename = self.sample_dir / self.get_sample_file_name(metadata)
        query_options = QueryOptions.from_metadata(metadata)
        needs_fresh_download = (not self.use_downloaded) or (not filename.is_file())

        if needs_fresh_download:
            logger.info("Performing request to TOAR DB")
            if filename.is_file():#delete old status endpoint, in case a fresh download is requested for an successful request.
                self.connection.cache.remove(query_options.cache_key)
            response = self.connection.get(query_options)
            with open(filename, "w+b") as downloaded_file:
                downloaded_file.write(response.content)
        else:
            logger.info(f"Loading already downloaded data to file {filename}")

        with open(filename, "r+b") as data_file:
            content = data_file.read()

        timeseries, timeseries_metadata = self.load_data(content, metadata)
        return timeseries, timeseries_metadata

    @staticmethod
    def get_sample_file_name(metadata: Metadata):
        """creates a filename from the metadata

        At the moment considering statistical method, sampling (temporal aggregation) as well as start and end.
        Parameters:
        ----------
        metadata:
            metadata for the request.
        """
        addition = "_".join(f"{key}{val}" for key, val in sorted(metadata.moreOptions.items()))
        addition = addition.replace("/", "%2F")
        return (
            "_".join(
                str(i)
                for i in [
                    metadata.statistic,
                    metadata.time.sampling,
                    metadata.variable.cf_standardname,
                    metadata.time.start.date(),
                    metadata.time.end.date(),
                    addition,
                ]
            )
            + ".zip"
        )
