import logging
from dataclasses import dataclass

import numpy as np
import xarray as xr

from toargridding.metadata import Metadata, Variables, get_cf_metadata

logger = logging.getLogger(__name__)


@dataclass
class Variable:
    """full variable including data and information according to CF

    CF: https://cfconventions.org/cf-conventions/cf-conventions.html

    Parameters:
    ----------
    var:
        name of the TOAR variable
    data:
        array with data
    attributes:
        metadata according to CF
    encoding:
        encoding of data type
    """

    var: Variables
    data: np.ndarray
    attributes: dict[str, str]
    encoding: dict[str, str]

    @classmethod
    def from_data(cls, data: np.array, variable: Variables, metadata: Metadata | None, **kwargs):
        """construction from analysis results"""
        cf_metadata = get_cf_metadata(variable, metadata=metadata)
        # print(variable.name, cf_metadata)
        return cls(variable, data, **cf_metadata, **kwargs)

    def as_data_array(self, dims) -> xr.DataArray:
        """conversion to DataArray"""
        da = xr.DataArray(
            self.data,
            name=self.name,
            dims=dims,
            attrs=self.attributes,
        )
        da.encoding = self.encoding
        return da

    @property
    def name(self):
        """shortcut to variable name"""
        return self.var.name

    @property
    def size(self):
        """shortcut to length of data array"""
        return self.data.size

    @property
    def min(self):
        """shortcut to minimum of data array"""
        return self.data.min()

    @property
    def max(self):
        """shortcut to maximum of data array"""
        return self.data.max()


@dataclass
class Coordinate(Variable):
    """coordinate axis"""

    step: float | str

    @classmethod
    def from_resolution(cls, variable: Variables, resolution: float, min: float, max: float, wraps: bool):
        """construction from a data range and resolution

        Creates a coordinate axis between min and amx with a step size close to resolution.

        Parameters:
        ----------
        resolution:
            width of a bin; actual size will be selected to obtain equidistant steps between steps
            If the resolution does not result in a set of equidistant bins, a bin is added and the resolution is slightly decreased.
        min:
            lowest value on coordinate axis
        max:
            highest value on coordinate axis
        wraps:
            if true, the last grid point is removed as it is the same as the first one, e.g. for longitude [-180, -180+res, 180-res].
            if false, the last grid point is kept like for the latitude [-90, -90+res... , 90-res,90]
        """

        span = max - min
        n = int(span / resolution)  # TODO: raise error if invalid inputs ?
        if n * resolution != span:
            logger.warning(
                f"Resolution {resolution} does not provide an equidistant division of the span [{min},{max}]"
            )
            n += 1
            step = span / n
            logger.warning(f"Adoption resolution {resolution} to {step}")
        else:
            step = resolution
        data = np.linspace(min, max, n + 1)

        if wraps:
            return cls.from_data(data[:-1], variable, None, step=step)
        else:
            return cls.from_data(data, variable, None, step=step)

    def as_data_array(self):
        """conversion to data array"""
        return super().as_data_array(dims=self.name)


def get_encoding(variables: list[Variable]) -> dict[str, dict]:
    """get dictionary encoding with encoding"""
    return {variable.name: variable.encoding for variable in variables}
